/*
 * Copyright (c) 2010 LAAS/CNRS
 * Arnaud Degroote - Wed Aug 11 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include <tclserv_client/buf.h>

buf_t buf_init()
{
	buf_t res = malloc(sizeof(*res));
	if (res == NULL)
		return NULL;

	res->data = NULL;
	res->len = 0;
	res->allocated = 0;

	return res;
}

void buf_destroy(buf_t* buf)
{
	if (buf == NULL || *buf == NULL)
		return;

	if ((*buf)->allocated)
		free((*buf)->data);
	free(*buf);
	*buf = NULL;
}

int buf_reserve(buf_t buf, size_t new_size)
{
	char* tmp;
	size_t allocated_size, next_size;
	int was_null = 0;

	if (buf->data == NULL) 
		was_null = 1;

	if (buf->allocated >= new_size)
		return 0;

	next_size = 2 * buf->allocated + 1;
	allocated_size = (next_size > new_size) ? next_size : new_size;

	tmp = realloc(buf->data, allocated_size);
	if (tmp == NULL)
		return -1;

	buf->data = tmp;
	if (was_null) 
		*(buf->data) = '\0';
	buf->allocated = allocated_size;
	return 0;
}

void
buf_cat(buf_t buf, const char *fmt, ...)
{
	char tmp_buf[256];
	va_list ap;

	va_start(ap, fmt);
	vsprintf(tmp_buf, fmt, ap);
	va_end(ap);

	buf_reserve(buf, buf->len + strlen(tmp_buf) + 1);
	strcat(buf->data, tmp_buf);
	buf->len = buf->len + strlen(tmp_buf);
}

// stolen from genom, will use standard definition when integrated
/*
 *  Macro pour parcourir les elements "elt" d'un tableau
 *  de "nDim" dimensions definies dans le tableau "dims" 
 *  Remarque: doit etre imperativement dans la suite immediate 
 *            de la partie de declarations de la fonction.
 */
#define FOR_EACH_elt(nDim,dim) \
{ int nElts=1,i,elt;\
	for (i=0;i<(nDim);i++) nElts *= (dims)[i];\
	for (elt=0;elt<nElts;elt++)

#define END_FOR }

void 
buf_add_char(buf_t buf, char* x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%c ", *(x+elt));
	END_FOR
}

void 
buf_add_unsigned_char(buf_t buf, unsigned char *x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%c ", *(x+elt));
	END_FOR
}

void
buf_add_int(buf_t buf, int* x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%d ", *(x+elt));
	END_FOR
}

void 
buf_add_long_int(buf_t buf, long int* x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%ld ", *(x+elt));
	END_FOR
}

void 
buf_add_long_long_int(buf_t buf, long long int* x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%lld ", *(x+elt));
	END_FOR
}

void 
buf_add_short_int(buf_t buf, short int* x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%d ", *(x+elt));
	END_FOR
}

void 
buf_add_unsigned_int(buf_t buf, unsigned int *x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%u ", *(x+elt));
	END_FOR
}

void 
buf_add_unsigned_long_int(buf_t buf, unsigned long *x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%lu ", *(x+elt));
	END_FOR
}

void 
buf_add_unsigned_long_long_int(buf_t buf, unsigned long long int* x, 
							    int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%llu ", *(x+elt));
	END_FOR
}

void 
buf_add_unsigned_short_int(buf_t buf, unsigned short *x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%u ", *(x+elt));
	END_FOR
}

void 
buf_add_double(buf_t buf, double* x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%lf ", *(x+elt));
	END_FOR
}
void
buf_add_float(buf_t buf, float* x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%f ", *(x+elt));
	END_FOR
}

void
buf_add_addr(buf_t buf, void** x, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims)
		buf_cat(buf, "%p ", *(x+elt));
	END_FOR
}

static void
buf_add_string_len(buf_t buf, char *x, int max_str_len, int nDim, int *dims)
{
	FOR_EACH_elt(nDim, dims) {
		char * p = x + elt * max_str_len;

		/* empty string must be encoded as "" */
		if (*p == '\0')
			p = "\"\"";

		size_t len = strlen(p);
		buf_reserve(buf, buf->len + len + 2);
		strcat(buf->data, p);
		strcat(buf->data, " ");
		buf->len += len + 1;
	} END_FOR
}

void
buf_add_string(buf_t buf, char *x, int nDim, int *dims)
{
	if (dims == NULL || nDim == 0) 
		return;

	if(dims[nDim-1] == 0) {
		fprintf(stderr, "Null string length\n");
		return;
	}

	buf_add_string_len(buf, x, dims[nDim-1], nDim-1, dims);
}
