/*
 * Copyright (c) 2010-2012 LAAS/CNRS
 * Arnaud Degroote - Wed Aug 11 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <pthread.h>

#include <tclserv_client/tclserv_client.h>
#include <tclserv_client/socket.h>

struct tclserv_socket
{
	tclserv_client_socket_t s;
	char* identifier;
	buf_t internal_buf;
};

enum state { INITIAL, ACK, TERM };

struct node
{
	struct node* next;
	int rqstId;
	int activityId;
	enum state state;
	char* answer;
};

typedef struct node* node_t;

struct tclserv_client 
{
	struct tclserv_socket* command;
	struct tclserv_socket* answer;

	node_t first;

	int abort;

	bool use_internal_thread;
	pthread_t thr;
	/* Protect critical section over list */
	pthread_mutex_t mtx_list;
	/* Protect critical section on condition variable */
	pthread_mutex_t mtx_cond;
	/* 
	 * Protect critical section over cmd socket => assure that we call
	 * sequentially write / read / treatment. Otherwise, some weird things can
	 * happen in multi thread context.
	 */
	pthread_mutex_t mtx_cmd;
	/* Signal changes in list */
	pthread_cond_t cond;
};

static struct tclserv_socket* tclserv_socket_init(const char* host, const char* port);
static void tclserv_socket_destroy(struct tclserv_socket** sock);
static void* tclserv_thread_main(void *);

static inline ssize_t
tclserv_client_write(TCLSERV_CLIENT_ID id, const char* buf)
{
	return tclserv_client_socket_write_request(id->command->s, buf);
}

static inline ssize_t
tclserv_client_read_command(TCLSERV_CLIENT_ID id)
{
	return tclserv_client_socket_read_request(id->command->s, id->command->internal_buf);
}

static inline ssize_t
tclserv_client_read_answer(TCLSERV_CLIENT_ID id)
{
	return tclserv_client_socket_read_request(id->answer->s, id->answer->internal_buf);
}

static inline char*
tclserv_client_get_error_msg(struct tclserv_socket* s)
{
	return (s->internal_buf->data + strlen("ERROR "));
}

static inline bool
is_ok(buf_t buf)
{
	char *p;
	if (buf == NULL || buf->data == NULL) 
		return false;

	p = strstr(buf->data, "OK");
	if (p == NULL || p != buf->data)
		return false;
	
	return true;
}

static struct tclserv_socket* 
tclserv_socket_init(const char* host, const char* port)
{
	struct tclserv_socket* res;
	char*p;
	size_t len;
	const char* hello_pattern = "HELLO ";
	size_t size_pattern = strlen(hello_pattern);

	res = malloc(sizeof(*res));
	if (res == NULL)
		goto clean2;

	res->s = tclserv_client_socket_init(host, port);
	if (res->s == NULL) 
		goto clean1;

	res->internal_buf = buf_init();
	if (res->internal_buf == NULL) 
		goto clean0;

	tclserv_client_socket_write_request(res->s, "HELLO\r\n");
	if (tclserv_client_socket_read_request(res->s, res->internal_buf) == -1)
		goto clean;

	p = strstr(res->internal_buf->data, hello_pattern);
	if (p == NULL || p != res->internal_buf->data)
		goto clean;

	len = (res->internal_buf->len - size_pattern - 2);
	res->identifier = malloc(len+1);
	if (res->identifier == NULL)
		goto clean;

	strncpy(res->identifier, res->internal_buf->data + size_pattern, len);
	res->identifier[len] = 0;
	return res;

clean:
	buf_destroy(&res->internal_buf);
clean0:
	tclserv_client_socket_destroy(&res->s);
clean1:
	free(res);
clean2:
	return NULL;
}

static void
tclserv_socket_destroy(struct tclserv_socket** sock)
{
	if (sock == NULL || *sock == NULL)
		return;
	free((*sock)->identifier);
	buf_destroy(&(*sock)->internal_buf);
	tclserv_client_socket_destroy(&(*sock)->s);
	free(*sock);
	*sock = NULL;
}

TCLSERV_CLIENT_ID
tclserv_client_init_ex(const char* host, const char* port, bool use_internal_thread)
{
	TCLSERV_CLIENT_ID res;
	char buf[256];

	res = malloc(sizeof(*res));
	if (res == NULL)
		goto clean7;

	res->command = tclserv_socket_init(host, port);
	if (res->command == NULL)
		goto clean6;

	res->answer = tclserv_socket_init(host, port);
	if (res->answer == NULL)
		goto clean5;

	// tell to the server to send answer on the answer socket
	snprintf(buf, sizeof(buf), "REPLYTO %s\r\n", res->answer->identifier);
	tclserv_client_write(res, buf);
	if (tclserv_client_read_command(res) == -1) {
		fprintf(stderr, "Can't setup answer canal\n");
		goto clean4;
	}

	// expected answer
	snprintf(buf, sizeof(buf), "OK %s\r\n", res->answer->identifier);
	if (strcmp(buf, res->command->internal_buf->data)) {
		fprintf(stderr, "can't setup answer canal : %s\n",
				res->command->internal_buf->data);
		goto clean4;
	}

	res->first = NULL;

	res->abort = 0;

	if (pthread_mutex_init(& res->mtx_cmd, NULL) == -1)
		goto clean4;
	if (pthread_mutex_init(& res->mtx_list, NULL) == -1)
		goto clean3;
	if (pthread_mutex_init(& res->mtx_cond, NULL) == -1)
		goto clean2;
	if (pthread_cond_init(& res->cond, NULL) == -1)
		goto clean1;
	res->use_internal_thread = use_internal_thread;
	if (use_internal_thread) {
		if (pthread_create(& res->thr, NULL, &tclserv_thread_main, res) == -1)
			goto clean;
	}

	return res;

clean:
	pthread_cond_destroy(& res->cond);
clean1:
	pthread_mutex_destroy(& res->mtx_cond);
clean2:
	pthread_mutex_destroy(& res->mtx_list);
clean3:
	pthread_mutex_destroy(& res->mtx_cmd);
clean4:
	tclserv_socket_destroy(&res->answer);
clean5:
	tclserv_socket_destroy(&res->command);
clean6:
	free(res);
clean7:
	return NULL;
}

TCLSERV_CLIENT_ID
tclserv_client_init(const char* host, const char* port)
{
	return tclserv_client_init_ex(host, port, true);
}

void
tclserv_client_destroy(TCLSERV_CLIENT_ID* id)
{
	if (id == NULL || *id == NULL)
		return;

	node_t next, tmp;

	(*id)->abort = 1;
	// send bye
	tclserv_client_socket_write_request((*id)->answer->s, "BYE\r\n");
	tclserv_client_socket_write_request((*id)->command->s, "BYE\r\n");

	if ((*id)->use_internal_thread)
		(void) pthread_join((*id)->thr, NULL);

	tclserv_socket_destroy(&(*id)->command);
	tclserv_socket_destroy(&(*id)->answer);


	// destroy the rqst list, don't need the mutex now
	next = (*id)->first;
	while (next != NULL) {
		tmp = next;
		next = next->next;
		if (tmp->answer != NULL)
			free(tmp->answer);
		free(tmp);
	}

	pthread_cond_destroy(& (*id)->cond);
	pthread_mutex_destroy(& (*id)->mtx_cond);
	pthread_mutex_destroy(& (*id)->mtx_list);
	pthread_mutex_destroy(& (*id)->mtx_cmd);

	free(*id);
	*id = NULL;
}

int tclserv_client_get_answer_socket(TCLSERV_CLIENT_ID id)
{
	return tclserv_client_socket_internal_descriptor(id->answer->s);
}

static inline int
send_command_and_check_ok(TCLSERV_CLIENT_ID id, const char* buf, char** err)
{
	pthread_mutex_lock(& id->mtx_cmd);
	tclserv_client_write(id, buf);
	if (tclserv_client_read_command(id) == -1) {
		fprintf(stderr, "Failed to call %s\n", buf);
		pthread_mutex_unlock(& id->mtx_cmd);
		return -1;
	}

	if (! is_ok(id->command->internal_buf)) {
		fprintf(stderr, "Failed to call %s : %s\n", buf, 
					id->command->internal_buf->data);

		if (err != NULL) 
			*err = strdup(tclserv_client_get_error_msg(id->command));
		pthread_mutex_unlock(& id->mtx_cmd);
		return -1;
	}

	pthread_mutex_unlock(& id->mtx_cmd);
	return 0;
}

static inline int
tclserv_client_free_rqstId(TCLSERV_CLIENT_ID id, ssize_t rqstId)
{
	char buf[32];

	snprintf(buf, sizeof(buf), "ACK %zd\r\n", rqstId);
	return send_command_and_check_ok(id, buf, NULL);
}

int
tclserv_client_load_module_ex(TCLSERV_CLIENT_ID id, const char* module, char** err)
{
	char buf[256];

	if (id == NULL || module == NULL)
		return -1;

	assert(strlen(module) < sizeof(buf) - 5); // 5 =  strlen("LM " + "\r\n")

	snprintf(buf, sizeof(buf), "LM %s\r\n", module);
	if (send_command_and_check_ok(id, buf, err))
		return -1;

	/*
	 * If cs::init failed, we try to disconnect all modules, generated bigger
	 * mbox, and reload all modules. If it fails, we can't do anything more, so
	 * badly fails
	 */
	pthread_mutex_lock(& id->mtx_cmd);
	tclserv_client_write(id, "cs::init\r\n");
	if (tclserv_client_read_command(id) == -1) {
		fprintf(stderr, "Failed to call %s\n", buf);
		pthread_mutex_unlock(& id->mtx_cmd);
		return -1;
	}

	if (! is_ok(id->command->internal_buf)) {
		if ( strstr(id->command->internal_buf->data, "too small") != NULL) {
			pthread_mutex_unlock(& id->mtx_cmd);
			if (send_command_and_check_ok(id, "modules::disconnect\r\n", err) == -1) 
				return -1;
			if (send_command_and_check_ok(id, "cs::exit\r\n", err) == -1)
				return -1;
			if (send_command_and_check_ok(id, "cs::init\r\n", err) == -1)
				return -1;
		} else {
			fprintf(stderr, "Failed to call %s : %s\n", buf, 
						id->command->internal_buf->data);
			pthread_mutex_unlock(& id->mtx_cmd);
			return -1;
		}
	} else  {
		pthread_mutex_unlock(& id->mtx_cmd);
	}

	if (send_command_and_check_ok(id, "modules::connect\r\n", err))
		return -1;

	return 0;
}

int
tclserv_client_load_module(TCLSERV_CLIENT_ID id, const char* module)
{
	return tclserv_client_load_module_ex(id, module, NULL);
}

int
tclserv_client_unload_module(TCLSERV_CLIENT_ID id, const char* module)
{
	char buf[256];

	if (id == NULL || module == NULL)
		return -1;

	assert(strlen(module) < sizeof(buf) - 7); // 7 strlen("UNLM " + "\r\n")
	snprintf(buf, sizeof(buf), "UNLM %s\r\n", module);
	if (send_command_and_check_ok(id, buf, NULL))
		return -1;

	return 0;
}

static ssize_t
tclserv_client_rpc(TCLSERV_CLIENT_ID id, const char* buf)
{
	int err;
	ssize_t rqstId;
	node_t new;

	/*
	 * Take the mutex early, otherwise the other thread can read the answer
	 * before we add the entry in the list
	 */
	pthread_mutex_lock(&id->mtx_list);
	pthread_mutex_lock(&id->mtx_cmd);
	tclserv_client_write(id, buf);

	if (tclserv_client_read_command(id) == -1) {
		fprintf(stderr, "can't get an answer for %s\n", buf);
		pthread_mutex_unlock(&id->mtx_cmd);
		pthread_mutex_unlock(&id->mtx_list);
		return -1;
	}

	err = sscanf(id->command->internal_buf->data, "OK %zd\r\n", &rqstId);
	if (err != 1) {
		fprintf(stderr, "bad answer for %s\n", buf);
		pthread_mutex_unlock(&id->mtx_cmd);
		pthread_mutex_unlock(&id->mtx_list);
		return -1;
	}
	pthread_mutex_unlock(&id->mtx_cmd);

	new = malloc(sizeof(*new));
	if (new == NULL) {
		fprintf(stderr, "Can't allocate memory for new node\n");
		pthread_mutex_unlock(&id->mtx_list);
		return -1;
	}
	
	new->state = INITIAL;
	new->answer = NULL;
	new->rqstId = rqstId;

	// insert in the head
	new->next = id->first;
	id->first = new;
	pthread_mutex_unlock(&id->mtx_list);

	return rqstId;
}

ssize_t
tclserv_client_rqst(TCLSERV_CLIENT_ID id, const char* rqst)
{
	if (id == NULL || rqst == NULL)
		return -1;

	char* buf;
	const char* rqst_pattern = "RQST ";
	const char* end_pattern = "\r\n";
	const size_t rqst_pattern_len = strlen(rqst_pattern);
	const size_t end_pattern_len = strlen(end_pattern);
	size_t len = strlen(rqst) + rqst_pattern_len + end_pattern_len;
	ssize_t rqstId;

	buf = malloc(len);
	if (buf == NULL)
		return -1;

	snprintf(buf, len, "%s%s%s", rqst_pattern, rqst, end_pattern);

	rqstId = tclserv_client_rpc(id, buf);

	free(buf);

	return rqstId;
}

int
tclserv_client_process_answer_msg(TCLSERV_CLIENT_ID id, ssize_t* pRqstId)
{
	return tclserv_client_timed_process_answer_msg(id, pRqstId, NULL);
}

int 
tclserv_client_timed_process_answer_msg(TCLSERV_CLIENT_ID id, ssize_t* pRqstId, struct timeval* tv)
{
	char buf[512];
	char *p, *begin;
	size_t rqstId;
	enum state new_state;
	node_t current;
	size_t len;
	int activityId;

	if (id == NULL)
		return -1;

	if (tclserv_client_socket_timed_read_request(id->answer->s, 
									id->answer->internal_buf, tv) < 0)
	{
		fprintf(stderr, "Error while reading answer channel\n");
		return -1;
	}

	// parse the answer
	if (id->answer->internal_buf->data == NULL) 
		return -1;

	// first get rqst id
	begin = id->answer->internal_buf->data;
	p = begin;
	while (*p != ' ' && *p != 0) p++;
	if (*p == 0) return -1;

	if ((p - begin) > sizeof(buf)) 
		return -1;
	memcpy(buf, begin, p - begin);
	buf[p - begin] = 0;

	if (sscanf(buf, "%zd", &rqstId) != 1) {
		fprintf(stderr, "First part of the answer is not a rqstId\n");
		return -1;
	}

	if (pRqstId != NULL)
		*pRqstId = rqstId;

	// then the command name (we don't care a lot)
	do { p++; } while (*p != ' ' && *p != 0);
	if (*p == 0) return -1;

	// then ACK or TERM
	p++;
	begin = p;
	while (*p != ' ' && *p != 0) p++;
	if (*p == 0) return -1;

	if ((p - begin) > sizeof(buf))
		return -1;
	memcpy(buf, begin, p - begin);
	buf[p - begin] = 0;

	if (strcmp(buf, "ACK") == 0) {
		new_state = ACK;
		p++;
		len = strlen(p) - 2; // for "/r/n"
		if (len > sizeof(buf))
			return -1;
		memcpy(buf, p, len);
		buf[len] = 0;
		if (sscanf(buf, "%d", &activityId) != 1)
			return -1;
	} else if (strcmp(buf, "TERM") == 0) {
		new_state = TERM;
	} else {
		fprintf(stderr, "Invalid state : %s\n", buf);
		return -1;
	}

	pthread_mutex_lock(&id->mtx_list);
	if (id->first == NULL) {
		fprintf(stderr, "NULL rqstId list : something is weird\n");
		pthread_mutex_unlock(&id->mtx_list);
		return -1;
	}

	current = id->first;
	while (current != NULL && current->rqstId != rqstId)
		current = current->next;

	if (current == NULL) {
		fprintf(stderr, "Can't find reference for rqstId %zd in the list\n", rqstId);
		pthread_mutex_unlock(&id->mtx_list);
		return -1;
	}

	pthread_mutex_lock(&id->mtx_cond);
	current->state = new_state;
	if (new_state == TERM) {
		current->answer = strdup(p+1);
		len = strlen(current->answer);
		current->answer[len - 2] = 0;
	} else {
		current->activityId = activityId;
	}
	pthread_cond_broadcast(&id->cond);
	pthread_mutex_unlock(&id->mtx_cond);
	pthread_mutex_unlock(&id->mtx_list);


	return 0;
}
   		
static void*
tclserv_thread_main(void * args)
{
	TCLSERV_CLIENT_ID id = args;
	while (! id->abort && 
		   tclserv_client_socket_is_connected(id->answer->s))
		tclserv_client_process_answer_msg(id, NULL);

	return NULL;
}

ssize_t 
tclserv_client_rqst_ack(TCLSERV_CLIENT_ID id, const char* rqst)
{
	ssize_t rqstId = tclserv_client_rqst(id, rqst);
	int get_ack = 0;
	node_t current = NULL;

	while (! get_ack) {
		if (current == NULL) {
			pthread_mutex_lock(& id->mtx_list );
			current = id->first;
			while (current != NULL && current->rqstId != rqstId)
				current = current->next;
			pthread_mutex_unlock( & id->mtx_list );
			if (current == NULL) // weird, not in the list
				return -1;
		}

		/* 
		 * we can keep current now, the only one we can remove it from the list
		 * is wait, but nobody will call in different thread ack and wait on
		 * the same rqstId
		 */

		pthread_mutex_lock(& id->mtx_cond);
		if (current->state == ACK || current->state == TERM) {
			get_ack = 1;
		} else {
			// wait until someone notifies some changes
			pthread_cond_wait(& id->cond, & id->mtx_cond );
		}
		pthread_mutex_unlock(& id->mtx_cond);
	}

	return rqstId;
}


int 
tclserv_client_wait(TCLSERV_CLIENT_ID id, ssize_t rqstId, char** msg)
{
	int get_term = 0;
	node_t current = NULL;
	node_t tmp = NULL;
	int err = -2;

	if (id == NULL || rqstId == -1)
		return -2;

	while (! get_term) {
		if (current == NULL) {
			pthread_mutex_lock(& id->mtx_list );
			current = id->first;
			while (current != NULL && current->rqstId != rqstId)
				current = current->next;
			pthread_mutex_unlock( & id->mtx_list );
			if (current == NULL) // weird, not in the list
				return -2;
		}

		// we can keep current now, we are the only one which can remove it
		// from the list

		pthread_mutex_lock(& id->mtx_cond);
		if (current->state == TERM) {
			get_term = 1;
		} else {
			// wait until someone notifies some changes
			pthread_cond_wait(& id->cond, &id->mtx_cond);
		}
		pthread_mutex_unlock(& id->mtx_cond);
	}

	if (current->answer[0] == 'O' && current->answer[1] == 'K')
		err = 0;
	else 
		err = -1;

	if (msg) {
		*msg = current->answer;
	} else {
		free(current->answer);
	}
 
	// removing the entry from the list, we don't care anymore
	pthread_mutex_lock(& id->mtx_list);
	if (id->first == current) {
		id->first = current->next;
	} else {
		tmp = id->first;
		while (tmp->next != current) 
			tmp = tmp->next;
		tmp->next = current->next;
	}
	free(current);
	pthread_mutex_unlock(& id->mtx_list);

	tclserv_client_free_rqstId(id, rqstId);

	return err;
}

int 
tclserv_client_rqst_wait(TCLSERV_CLIENT_ID id, const char* rqst, char** msg)
{
	ssize_t rqstId = tclserv_client_rqst(id, rqst);
	return tclserv_client_wait(id, rqstId, msg);
}

int 
tclserv_client_has_terminated(TCLSERV_CLIENT_ID id, ssize_t rqstId)
{
	node_t current = NULL;

	pthread_mutex_lock(& id->mtx_list );
	current = id->first;
	while (current != NULL && current->rqstId != rqstId)
		current = current->next;
	pthread_mutex_unlock( & id->mtx_list );
	if (current == NULL) // weird, not in the list
		return -2;

	if (current->state == TERM) 
		return 0;

	return -1;
}

ssize_t
tclserv_client_abort_ex(TCLSERV_CLIENT_ID id, ssize_t rqstId, bool wait)
{
	node_t current = NULL;
	char buf[256];

	if (id == NULL || rqstId < 0)
		return -2;

	pthread_mutex_lock(& id->mtx_list );
	current = id->first;
	while (current != NULL && current->rqstId != rqstId)
		current = current->next;
	pthread_mutex_unlock( & id->mtx_list );
	if (current == NULL) // weird, not in the list
		return -2;

	if (current->state == TERM) {
		fprintf(stderr, "rqst already terminated, no point to abort it\n");
		return -2;
	}

	if (current->state == INITIAL) {
		// wait until the ACK state
		if (wait) {
			pthread_mutex_lock(& id->mtx_cond);

			while (current->state != ACK ) {
				pthread_cond_wait(& id->cond, &id->mtx_cond);
			}
			pthread_mutex_unlock(& id->mtx_cond);
		} else {
			return -3;
		}
	}

	if (current->state == ACK) {
		snprintf(buf, sizeof(buf), "ABORT %d\r\n", current->rqstId);
		ssize_t abortRqstId = tclserv_client_rpc(id, buf);
		if (wait) 
			return tclserv_client_wait(id, abortRqstId, NULL);
		else 
			return abortRqstId;
	}

	return -2; // invalid state :D
}

int 
tclserv_client_abort(TCLSERV_CLIENT_ID id, ssize_t rqstId)
{
	tclserv_client_abort_ex(id, rqstId, true);
}

int
tclserv_client_list_modules(TCLSERV_CLIENT_ID id, char** answer)
{
	char* p;
	int err = -2;

	if (id == NULL || answer == NULL)
		return -2;

	pthread_mutex_lock(&id->mtx_cmd);
	if (tclserv_client_write(id, "cs::lsmbox\r\n") < 0) {
		fprintf(stderr, "Failed to send LSMBOX request\n");
		pthread_mutex_unlock(&id->mtx_cmd);
		return -2;
	}

	if (tclserv_client_read_command(id) < 0 ) {
		fprintf(stderr, "Can't reed the answer of LSMBOX answer\n");
		pthread_mutex_unlock(&id->mtx_cmd);
		return -2;
	}

	// check the answer stored in id->command->internal_buf
	p = strstr(id->command->internal_buf->data, "OK ");
	if ( p == NULL || p != id->command->internal_buf->data ) {
		*answer = strdup(id->command->internal_buf->data);
		err = -1;
	} else {
		*answer = strdup(id->command->internal_buf->data + 3); // "OK "
		err = 0;
	}
	pthread_mutex_unlock(&id->mtx_cmd);
	return err;
}

int 
tclserv_client_exec(TCLSERV_CLIENT_ID id, const char* rqst, char** answer)
{
	char* buf;
	char* p;
	int err = -2;

	if (id == NULL || rqst == NULL)
		return -2;

	buf = malloc(strlen(rqst) + 12); // 12 is enough to store "exec --" + "\r\n"
	if (buf == NULL)
		return -2;

	strcpy(buf, "exec -- ");
	strcat(buf, rqst);
	strcat(buf, "\r\n");

	pthread_mutex_lock(&id->mtx_cmd);
	if (tclserv_client_write(id, buf) < 0)  {
		fprintf(stderr, "Failed to send %s request\n", buf);
		pthread_mutex_unlock(&id->mtx_cmd);
		free(buf);
		return -2;
	}

	if (tclserv_client_read_command(id) < 0) {
		fprintf(stderr, "Can't read the answer of %s request\n", buf);
		pthread_mutex_unlock(&id->mtx_cmd);
		free(buf);
		return -2;
	}

	free(buf);
	p = strstr(id->command->internal_buf->data , "OK ");
	if (p == NULL || p != id->command->internal_buf->data)  // error case
	{
		if (answer != NULL) 
			*answer = strdup(tclserv_client_get_error_msg(id->command));
		err = -1;
	} else {
		if (answer != NULL)
			*answer = strdup(p + 3); 
		err = 0;
	}

	pthread_mutex_unlock(&id->mtx_cmd);
	return err;
}
int 
tclserv_client_start_module(TCLSERV_CLIENT_ID id, const char* module)
{
	return tclserv_client_start_module_ex(id, module, NULL);
}

int 
tclserv_client_start_module_ex(TCLSERV_CLIENT_ID id, const char* module, char** why)
{
	char buf[256];
	/* Always provide a why buffer, to get the answer, and compute the pid */
	char* intern_why;
	char** cur_why = (why != NULL) ? why : &intern_why;

	if (id == NULL || module == NULL)
		return -2;

	if (strlen(module) > sizeof(buf) - 5) // 3 == strlen(" -b &")
		return -2;
	snprintf(buf, sizeof(buf), "%s -b &", module);
	
	int err = tclserv_client_exec(id, buf, cur_why);
	if (err < 0) {
		if (cur_why != why) 
			free(*cur_why);
		return err;
	}

	// extract pid from why
	sscanf(*cur_why, "%d\n", &err);
	free(*cur_why);
	return err;
}

int
tclserv_client_stop_module(TCLSERV_CLIENT_ID id, const char* module)
{
	return tclserv_client_stop_module_ex(id, module, NULL);
}

int
tclserv_client_stop_module_ex(TCLSERV_CLIENT_ID id, const char* module, char **why)
{
	char buf[256];
	int err;
	ssize_t rqstId;

	if (id == NULL || module == NULL)
		return -2;

	if (strlen(module) > sizeof(buf) - 13) // 13 == strlen(KILL \r\n) + strlen( " abort")
		return -2;
	snprintf(buf, sizeof(buf), "KILL %s abort\r\n", module);
	rqstId = tclserv_client_rpc(id, buf);
	if (rqstId < 0) 
		return -1;
	return tclserv_client_wait(id, rqstId, why);
}


